from .utils import Frame


class BaseEditor(Frame):
    """
    Base class for editors.
    """
    def __init__(self, master, path=None, path2=None, editable=True, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.background)

        self.path = path
        self.path2 = path2
        self.editable = editable

        self.showpath = False
        self.content = None
        self.diff = False

    def save(self, *_):
        ...



import re, codecs, os
import threading, queue
import tkinter as tk
from collections import deque

from .syntax import Syntax
from .highlighter import Highlighter
from .autocomplete import AutoComplete

from ..utils import Text


class Text(Text):
    def __init__(self, master, path=None, minimalist=False, language=None, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.path = path
        self.data = None
        self.encoding = 'utf-8'
        self.minimalist = minimalist
        
        self.buffer_size = 1000
        self.bom = True
        self.current_word = None
        self.words = []

        if self.path and os.path.isfile(self.path):
            self.load_file()

        self.syntax = Syntax(self)
        self.auto_completion = AutoComplete(
            self, items=self.syntax.get_autocomplete_list()) if not minimalist else None
        
        self.highlighter = Highlighter(self, language)

        self.focus_set()
        self.config_tags()
        self.create_proxy()
        self.config_bindings()
        self.configure(wrap=tk.NONE, relief=tk.FLAT, bg=self.base.theme.background, fg=self.base.theme.foreground)

        self.update_words()

    def config_tags(self):
        self.tag_config(tk.SEL, background=self.base.theme.editor.selection)
        self.tag_config("highlight", background=self.base.theme.editor.currentword)
        self.tag_config("currentline", background=self.base.theme.editor.currentline)

        self.tag_config("found", background=self.base.theme.editor.found)
        self.tag_config("foundcurrent", background=self.base.theme.editor.foundcurrent)

    def config_bindings(self):
        self.bind("<KeyRelease>", self.key_release_events) 

        self.bind("<Control-f>", self.open_find_replace)
        self.bind("<Control-d>", self.multi_selection)
        self.bind("<Control-Left>", lambda e: self.handle_ctrl_hmovement())
        self.bind("<Control-Right>", lambda e: self.handle_ctrl_hmovement(True))

        self.bind("<Return>", self.enter_key_events)
        self.bind("<Tab>", self.tab_key_events)

        if self.minimalist:
            return

        self.bind("<FocusOut>", self.hide_autocomplete) 
        self.bind("<Button-1>", self.hide_autocomplete)
        self.bind("<Up>", self.auto_completion.move_up)
        self.bind("<Down>", self.auto_completion.move_down)

        # self.bind("<braceleft>", lambda e: self.complete_pair("}"))
        # self.bind("<bracketleft>", lambda e: self.complete_pair("]"))
        # self.bind("<parenleft>", lambda e: self.complete_pair(")"))

        # self.bind("<apostrophe>", lambda e: self.surrounding_selection("\'"))
        # self.bind("<quotedbl>", lambda e: self.surrounding_selection("\""))

    def key_release_events(self, event):
        if event.keysym not in ("Up", "Down", "Return"):
            self.show_autocomplete(event)

        match event.keysym:
            # autocompletion keys
            case "Button-2" | "BackSpace" | "Escape" | "Control_L" | "Control_R" | "space":
                self.hide_autocomplete()
            case "rightarrow" | "leftarrow":
                self.update_completions()
            
            # bracket pair completions
            case "braceleft":
                return self.complete_pair("}")
            case "bracketleft":
                return self.complete_pair("]")
            case "parenleft":
                return self.complete_pair(")")

            # surroundings for selection
            case "apostrophe":
                return self.surrounding_selection("\'")
            case "quotedbl":
                return self.surrounding_selection("\"")

            # extra spaces
            case ":" | ",":
                self.insert(tk.INSERT, " ")

            case _:
                pass

    def enter_key_events(self, *_):
        if not self.minimalist and self.auto_completion.active:        
            self.auto_completion.choose()
            return "break"
        
        return self.check_indentation()
        
    def tab_key_events(self, *_):
        if self.auto_completion.active:        
            self.auto_completion.choose()
            return "break"
    
        self.insert(tk.INSERT, " "*4)
        return "break"
    
    def get_all_text(self, *args):
        return self.get(1.0, tk.END)

    def get_all_text_ac(self, *args):
        """
        Helper function for autocomplete.show
        extracts all text except the current word.
        """
        return self.get(1.0, "insert-1c wordstart-1c") + self.get("insert+1c", tk.END)
    
    def get_current_word(self):
        return self.current_word.strip()
    
    def update_words(self, *_):
        if self.minimalist:
            return
        
        self.words = list(set(re.findall(r"\w+", self.get_all_text_ac())))
        self.after(1000, self.update_words)
    
    def update_completions(self):
        if self.minimalist:
            return
        
        self.auto_completion.update_completions()   
    
    def confirm_autocomplete(self, text):
        self.replace_current_word(text)
        
    def replace_current_word(self, new_word):
        if self.current_word.startswith("\n"):
            self.delete("insert-1c wordstart+1c", "insert")
        else:
            self.delete("insert-1c wordstart", "insert")
        self.insert("insert", new_word)
    
    def check_autocomplete_keys(self, event):
        """
        Helper function for autocomplete.show to check triggers
        """
        return True if event.keysym not in [
            "BackSpace", "Escape", "Return", "Tab", "space", 
            "Up", "Down", "Control_L", "Control_R"] else False 
    
    def cursor_screen_location(self):
        """
        Helper function for autocomplete.show to detect cursor location
        """
        pos_x, pos_y = self.winfo_rootx(), self.winfo_rooty()

        cursor = tk.INSERT
        bbox = self.bbox(cursor)
        if not bbox:
            return (0, 0)
        
        bbx_x, bbx_y, _, bbx_h = bbox
        return (pos_x + bbx_x - 1, pos_y + bbx_y + bbx_h)
    
    def hide_autocomplete(self, *_):
        if self.minimalist:
            return
        
        self.auto_completion.hide()
    
    def show_autocomplete(self, event):
        if self.minimalist or not self.check_autocomplete_keys(event):
            return

        if self.current_word.strip() not in ["{", "}", ":", "", None, "\""] and not self.current_word.strip()[0].isdigit():
            if not self.auto_completion.active:
                if event.keysym in ["Left", "Right"]:
                    return
                pos = self.cursor_screen_location()
                self.auto_completion.show(pos)
                self.auto_completion.update_completions()
            else:
                self.auto_completion.update_completions()
        else:
            if self.auto_completion.active:
                self.hide_autocomplete()

    def complete_pair(self, char):
        self.insert(tk.INSERT, char)
        self.mark_set(tk.INSERT, "insert-1c")
    
    def surrounding_selection(self, char):
        next_char = self.get("insert", "insert+1c")
        if next_char == char:
            self.mark_set(tk.INSERT, "insert+1c")
            self.delete("insert-1c", "insert")
            return "break"

        if self.tag_ranges(tk.SEL):
            self.insert(char, tk.SEL_LAST)
            self.insert(char, tk.SEL_FIRST)
            return
        
        self.complete_pair(char)
        return "break"

    def move_to_next_word(self):
        self.mark_set(tk.INSERT, self.index("insert+1c wordend"))

    def move_to_previous_word(self):
        self.mark_set(tk.INSERT, self.index("insert-1c wordstart"))

    def handle_ctrl_hmovement(self, delta=False):
        if delta:
            self.move_to_next_word()
        else:
            self.move_to_previous_word()
        
        return "break"

    def update_current_indent(self):
        line = self.get("insert linestart", "insert lineend")
        match = re.match(r'^(\s+)', line)
        self.current_indent = len(match.group(0)) if match else 0

    def update_current_line(self):
        self.current_line = self.get("insert linestart", "insert lineend")
        return self.current_line
    
    def add_newline(self, count=1):
        self.insert(tk.INSERT, "\n" * count)
    
    def check_indentation(self, *args):
        self.update_current_indent()
        if self.update_current_line():
            if self.current_line[-1] in ["{", "[", ":", "("]:
                self.current_indent += 4
            elif self.current_line[-1] in ["}", "]", ")"]:
                self.current_indent -= 4
            
            self.add_newline()
            self.insert(tk.INSERT, " " * self.current_indent)

            self.update_current_indent()
            
            return "break"

    def multi_selection(self, *args):
        #TODO: multi cursor editing
        return "break"

    def open_find_replace(self, *_):
        self.base.findreplace.show(self)

    def detect_encoding(self, file_path):
        with open(file_path, 'rb') as file:
            bom = file.read(4)

        if bom.startswith(codecs.BOM_UTF8):
            return 'utf-8'
        elif bom.startswith(codecs.BOM_LE) or bom.startswith(codecs.BOM_BE):
            return 'utf-16'
        elif bom.startswith(codecs.BOM32_BE) or bom.startswith(codecs.BOM32_LE):
            return 'utf-32'

        self.bom = False
        return 'utf-8'

    def load_file(self):
        try:
            encoding = self.detect_encoding(self.path)
            file = open(self.path, 'r', encoding=encoding)
            self.encoding = encoding

            self.queue = queue.Queue()
            threading.Thread(target=self.read_file, args=(file,)).start()
            self.process_queue()
        except Exception as e:
            self.master.unsupported_file()

    def read_file(self, file):
        while True:
            chunk = file.read(self.buffer_size)
            if not chunk:
                file.close()
                self.queue.put(None)  # Signal the end of reading
                break
            self.queue.put(chunk)

    def process_queue(self):
        try:
            while True:
                chunk = self.queue.get_nowait()
                if chunk is None:
                    self.master.on_change()
                    self.master.on_scroll()
                    break
                self.write(chunk)
                self.update()
                self.master.on_scroll()
        except queue.Empty:
            # If the queue is empty, schedule the next check after a short delay
            self.master.after(100, self.process_queue)
    
    def save_file(self, path=None):
        if path:
            try:
                with open(path, 'w') as fp:
                    fp.write(self.get_all_text())
            except Exception:
                return
            
            self.path = path
        try:
            with open(self.path, 'w') as fp:
                fp.write(self.get_all_text())
        except Exception:
            return
             
    def copy(self, *_):
        self.event_generate("<<Copy>>")

    def cut(self, *_):
        self.event_generate("<<Cut>>")
    
    def paste(self, *_):
        self.event_generate("<<Paste>>")

    def set_data(self, data):
        self.data = data
        
    def clear(self):
        self.delete(1.0, tk.END)

    def write(self, text, *args):
        self.insert(tk.END, text, *args)
    
    def newline(self, *args):
        self.write("\n", *args)
    
    def get_all_text(self):
        return self.get(1.0, tk.END)
    
    def get_selected_text(self):
        try:
            return self.selection_get()
        except Exception:
            return ""
        
    def add_newline(self, count=1):
        self.insert(tk.INSERT, "\n" * count)

    def get_selected_count(self):
        return len(self.get_selected_text())
        
    @property
    def line(self):
        return int(self.index(tk.INSERT).split('.')[0])
    
    @property
    def column(self):
        return int(self.index(tk.INSERT).split('.')[1]) + 1

    @property
    def position(self):
        lc = self.index(tk.INSERT).split('.')
        return [lc[0], int(lc[1]) + 1]

    def scroll_to_end(self):
        self.mark_set(tk.INSERT, tk.END)
        self.see(tk.INSERT)
    
    def scroll_to_start(self):
        self.mark_set(tk.INSERT, 1.0)
        self.see(tk.INSERT)
    
    def scroll_to_line(self, line):
        self.mark_set(tk.INSERT, line)
        self.see(tk.INSERT)
    
    def set_wrap(self, flag=True):
        if flag:
            self.configure(wrap=tk.WORD)
        else:
            self.configure(wrap=tk.NONE)
    
    def set_active(self, flag=True):
        if flag:
            self.configure(state=tk.NORMAL)
        else:
            self.configure(state=tk.DISABLED)
    
    def show_unsupported_dialog(self):
        self.set_wrap(True)
        self.configure(font=('Arial', 10), padx=10, pady=10)
        self.write("This file is not displayed in this editor because it is either binary or uses an unsupported text encoding.")
        self.set_active(False)

    def move_cursor(self, position):
        self.mark_set(tk.INSERT, position)

    def clear_all_selection(self):
        self.tag_remove(tk.SEL, 1.0, tk.END)

    def highlight_current_line(self, *_):
        self.tag_remove("currentline", 1.0, tk.END)
        if self.get_selected_text():
            return
        
        line = int(self.index(tk.INSERT).split(".")[0])
        start = str(float(line))
        end = str(float(line + 1))
        self.tag_add("currentline", start, end)

    def select_line(self, line):
        self.clear_all_selection()
        
        line = int(line.split(".")[0])
        start = str(float(line))
        end = str(float(line + 1))
        self.tag_add(tk.SEL, start, end)

        self.move_cursor(end)
    
    def highlight_current_word(self):
        if self.minimalist or self.get_selected_text():
            return

        self.tag_remove("highlight", 1.0, tk.END)
        word = re.findall(r"\w+", self.get("insert wordstart", "insert wordend"))
        if any(word) and word[0] not in self.syntax.keywords:
            self.highlight_pattern(f"\\y{word[0]}\\y", "highlight", regexp=True)

        # elif word := self.get_selected_text():
        #     self.highlight_pattern(word, "highlight", end="sel.first")
        #     self.highlight_pattern(word, start="sel.last")


    def highlight_pattern(self, pattern, tag, start="1.0", end=tk.END, regexp=False):
        start = self.index(start)
        end = self.index(end)
        
        self.mark_set("matchStart", start)
        self.mark_set("matchEnd", start)
        self.mark_set("searchLimit", end)

        self.tag_remove(tag, start, end)
        
        count = tk.IntVar()
        while True:
            index = self.search(pattern, "matchEnd", "searchLimit", count=count, regexp=regexp)
            if index == "" or count.get() == 0:
                break

            self.mark_set("matchStart", index)
            self.mark_set("matchEnd", f"{index}+{count.get()}c")

            self.tag_add(tag, "matchStart", "matchEnd")

    def refresh(self, *args):
        if self.minimalist:
            return
        
        self.current_word = self.get("insert-1c wordstart", "insert")
        self.highlighter.highlight()
        self.highlight_current_line()
        self.highlight_current_word()

    def create_proxy(self):
        self._orig = self._w + "_orig"
        self.tk.call("rename", self._w, self._orig)
        self.tk.createcommand(self._w, self._proxy)

    def _proxy(self, *args):
        if args[0] == 'get' and (args[1] == tk.SEL_FIRST and args[2] == tk.SEL_LAST) and not self.tag_ranges(tk.SEL): 
            return
        if args[0] == 'delete' and (args[1] == tk.SEL_FIRST and args[2] == tk.SEL_LAST) and not self.tag_ranges(tk.SEL): 
            return

        cmd = (self._orig,) + args
        result = self.tk.call(cmd)

        if (args[0] in ("insert", "replace", "delete") or args[0:3] == ("mark", "set", "insert")):
            self.event_generate("<<Change>>", when="tail")
        
        elif (args[0:2] == ("xview", "moveto") or args[0:2] == ("yview", "moveto") or 
              args[0:2] == ("xview", "scroll") or args[0:2] == ("yview", "scroll")):
            self.event_generate("<<Scroll>>", when="tail")
            
        return result



import tkinter as tk


class Breakpoint(tk.Label):
    def __init__(self, master, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.base = master.base

        self.config(text="●", font=("Consolas", 14), fg="#1e1e1e", cursor="hand2", 
                    bg=self.base.theme.linenumbers["background"], borderwidth=0, 
                    width=2, height=1, pady=0, padx=0, relief=tk.FLAT)

        self.active = False
        self.hovered = False
        self.config_bindings()
    
    def config_bindings(self):
        self.bind("<Button-1>", self.on_click)
        self.bind("<Enter>", self.on_enter)
        self.bind("<Leave>", self.on_leave)

    def redraw(self):
        if self.active:
            self.config(fg="#e51400")
            return 
        
        if self.hovered:
            self.config(fg="#6e1911")
        else:
            self.config(fg="#1e1e1e")
        
    def on_click(self, event):
        self.active = not self.active
        self.redraw()
    
    def on_enter(self, event):
        self.hovered = True
        self.redraw()

    def on_leave(self, event):
        self.hovered = False
        self.redraw()



import os, tkinter as tk

from pygments import lex
from pygments.lexers import get_lexer_for_filename, get_lexer_by_name


class Highlighter:
    def __init__(self, master, language=None, *args, **kwargs):
        self.text = master
        self.base = master.base
        self.language = language
        
        if language:
            try:
                self.lexer = get_lexer_by_name(language)
            except:
                self.lexer = None
                return
        else:
            try:
                self.lexer = get_lexer_for_filename(os.path.basename(master.path), inencoding=master.encoding, encoding=master.encoding)
            except:
                self.lexer = None
                return

        self.setup_highlight_tags()

    def setup_highlight_tags(self):
        for token, color in self.base.settings.syntax.items():
            self.text.tag_configure(f"Token.{token}", foreground=color)

    def highlight(self):
        if not self.lexer:
            return
        
        for token, _ in self.base.settings.syntax.items():
            self.text.tag_remove(f"Token.{token}", '1.0', tk.END)
            
        text = self.text.get_all_text()

        # NOTE:  Highlighting only visible area
        # total_lines = int(self.text.index('end-1c').split('.')[0])
        # start_line = int(self.text.yview()[0] * total_lines)
        # first_visible_index = f"{start_line}.0"
        # last_visible_index =f"{self.text.winfo_height()}.end"
        # for token, _ in self.tag_colors.items():
        #     self.text.tag_remove(str(token), first_visible_index, last_visible_index)
        # text = self.text.get(first_visible_index, last_visible_index)

        self.text.mark_set("range_start", '1.0')
        for token, content in lex(text, self.lexer):
            self.text.mark_set("range_end", f"range_start + {len(content)}c")
            self.text.tag_add(str(token), "range_start", "range_end")
            self.text.mark_set("range_start", "range_end")
            
            # DEBUG
            # print(f"{content} is recognized as a <{str(token)}>")
        # print("==================================")



import tkinter as tk

from PIL import Image, ImageTk
from ..editor import BaseEditor


#TODO: zooming in and out
class ImageViewer(BaseEditor):
    def __init__(self, master, path, *args, **kwargs):
        super().__init__(master, path, editable=False, *args, **kwargs)
        self.open_image()

    def open_image(self):
        self.image = Image.open(self.path)
        self.image.thumbnail((700, 700))
        self.tk_image = ImageTk.PhotoImage(self.image)

        self.image_label = tk.Label(self, image=self.tk_image, bg=self.base.theme.background)
        self.image_label.pack(fill=tk.BOTH, expand=True)



import tkinter.ttk as ttk
import tkinter as tk

from .scrollbar import Scrollbar
from .frame import Frame


class Tree(Frame):
    def __init__(self, master, path=None, doubleclick=lambda _: None, singleclick=lambda _: None, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.tree["background"])

        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)

        self.path = path
        self.doubleclick = doubleclick
        self.singleclick = singleclick

        self.tree = ttk.Treeview(self, show="tree", columns=("fullpath", "type"), 
                                 displaycolumns='', selectmode=tk.BROWSE)
        self.tree.grid(row=0, column=0, sticky=tk.NSEW)
        
        self.scrollbar = Scrollbar(self, orient=tk.VERTICAL, command=self.tree.yview, style="TreeScrollbar")
        self.scrollbar.grid(row=0, column=1, sticky=tk.NS)

        self.tree.config(yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.tree.yview)
        
        self.tree.bind("<Double-Button-1>", self.doubleclick)
        self.tree.bind("<<TreeviewSelect>>", self.check_singleclick)

    def check_singleclick(self, _):
        if self.item_type(self.focus()) == 'file':
            if self.singleclick:
                self.singleclick(self.item_fullpath(self.focus()))
        else:
            self.toggle_node(self.focus())
    
    def clear_node(self, node):
        self.tree.delete(*self.tree.get_children(node))

    def clear_tree(self):
        self.clear_node('')

    def collapse_all(self):
        for node in self.tree.get_children():
            self.tree.item(node, open=False)
    
    def delete(self, *a, **kw):
        self.tree.delete(*a, *kw)
    
    def focus(self):
        return self.tree.focus()

    def get_children(self, *a, **kw):
        return self.tree.get_children(*a, **kw)

    def insert(self, *args, **kwargs):
        return self.tree.insert(*args, **kwargs)
        
    def is_open(self, node):
        return self.tree.item(node, "open")
    
    def item(self, *a, **kw):
        return self.tree.item(*a, **kw)
        
    def item_type(self, item):
        return self.set(item, "type")
    
    def item_fullpath(self, item):
        return self.set(item, "fullpath")

    def selected_path(self):
        return self.item_fullpath(self.focus())

    def selected_type(self):
        return self.item_type(self.focus())

    def set(self, *args, **kwargs):
        return self.tree.set(*args, **kwargs)

    def toggle_node(self, node):
        if self.item_type(node) == 'directory':
            if self.is_open(node):
                self.tree.item(node, open=False)
            else:
                self.tree.item(node, open=True)



class Languages:
    """
    Holds all available languages. To be passed to Editor during initialization.
    """
    ABAP = "abap"
    AMDGPU = "amdgpu"
    APL = "apl"
    ABNF = "abnf"
    ACTIONSCRIPT3 = "actionscript3"
    ACTIONSCRIPT = "actionscript"
    ADA = "ada"
    ADL = "adl"
    AGDA = "agda"
    AHEUI = "aheui"
    ALLOY = "alloy"
    AMBIENTTALK = "ambienttalk"
    AMPL = "ampl"
    HTML_NG2 = "html+ng2"
    NG2 = "ng2"
    ANTLR_ACTIONSCRIPT = "antlr-actionscript"
    ANTLR_CSHARP = "antlr-csharp"
    ANTLR_CPP = "antlr-cpp"
    ANTLR_JAVA = "antlr-java"
    ANTLR = "antlr"
    ANTLR_OBJC = "antlr-objc"
    ANTLR_PERL = "antlr-perl"
    ANTLR_PYTHON = "antlr-python"
    ANTLR_RUBY = "antlr-ruby"
    APACHECONF = "apacheconf"
    APPLESCRIPT = "applescript"
    ARDUINO = "arduino"
    ARROW = "arrow"
    ARTURO = "arturo"
    ASC = "asc"
    ASPECTJ = "aspectj"
    ASYMPTOTE = "asymptote"
    AUGEAS = "augeas"
    AUTOIT = "autoit"
    AUTOHOTKEY = "autohotkey"
    AWK = "awk"
    BBCBASIC = "bbcbasic"
    BBCODE = "bbcode"
    BC = "bc"
    BST = "bst"
    BARE = "bare"
    BASEMAKE = "basemake"
    BASH = "bash"
    CONSOLE = "console"
    BATCH = "batch"
    BDD = "bdd"
    BEFUNGE = "befunge"
    BERRY = "berry"
    BIBTEX = "bibtex"
    BLITZBASIC = "blitzbasic"
    BLITZMAX = "blitzmax"
    BNF = "bnf"
    BOA = "boa"
    BOO = "boo"
    BOOGIE = "boogie"
    BRAINFUCK = "brainfuck"
    BUGS = "bugs"
    CAMKES = "camkes"
    C = "c"
    CMAKE = "cmake"
    C_OBJDUMP = "c-objdump"
    CPSA = "cpsa"
    CSS_UL4 = "css+ul4"
    ASPX_CS = "aspx-cs"
    CSHARP = "csharp"
    CA65 = "ca65"
    CADL = "cadl"
    CAPDL = "capdl"
    CAPNP = "capnp"
    CBMBAS = "cbmbas"
    CDDL = "cddl"
    CEYLON = "ceylon"
    CFENGINE3 = "cfengine3"
    CHAISCRIPT = "chaiscript"
    CHAPEL = "chapel"
    CHARMCI = "charmci"
    HTML_CHEETAH = "html+cheetah"
    JAVASCRIPT_CHEETAH = "javascript+cheetah"
    CHEETAH = "cheetah"
    XML_CHEETAH = "xml+cheetah"
    CIRRU = "cirru"
    CLAY = "clay"
    CLEAN = "clean"
    CLOJURE = "clojure"
    CLOJURESCRIPT = "clojurescript"
    COBOLFREE = "cobolfree"
    COBOL = "cobol"
    COFFEESCRIPT = "coffeescript"
    CFC = "cfc"
    CFM = "cfm"
    CFS = "cfs"
    COMAL = "comal"
    COMMON_LISP = "common-lisp"
    COMPONENTPASCAL = "componentpascal"
    COQ = "coq"
    CPLINT = "cplint"
    CPP = "cpp"
    CPP_OBJDUMP = "cpp-objdump"
    CRMSH = "crmsh"
    CROC = "croc"
    CRYPTOL = "cryptol"
    CR = "cr"
    CSOUND_DOCUMENT = "csound-document"
    CSOUND = "csound"
    CSOUND_SCORE = "csound-score"
    CSS_DJANGO = "css+django"
    CSS_RUBY = "css+ruby"
    CSS_GENSHITEXT = "css+genshitext"
    CSS = "css"
    CSS_PHP = "css+php"
    CSS_SMARTY = "css+smarty"
    CUDA = "cuda"
    CYPHER = "cypher"
    CYTHON = "cython"
    D = "d"
    D_OBJDUMP = "d-objdump"
    DPATCH = "dpatch"
    DART = "dart"
    DASM16 = "dasm16"
    DEBCONTROL = "debcontrol"
    DELPHI = "delphi"
    DEVICETREE = "devicetree"
    DG = "dg"
    DIFF = "diff"
    DJANGO = "django"
    DOCKER = "docker"
    DTD = "dtd"
    DUEL = "duel"
    DYLAN_CONSOLE = "dylan-console"
    DYLAN = "dylan"
    DYLAN_LID = "dylan-lid"
    ECL = "ecl"
    EC = "ec"
    EARL_GREY = "earl-grey"
    EASYTRIEVE = "easytrieve"
    EBNF = "ebnf"
    EIFFEL = "eiffel"
    IEX = "iex"
    ELIXIR = "elixir"
    ELM = "elm"
    ELPI = "elpi"
    EMACS_LISP = "emacs-lisp"
    EMAIL = "email"
    ERB = "erb"
    ERLANG = "erlang"
    ERL = "erl"
    HTML_EVOQUE = "html+evoque"
    EVOQUE = "evoque"
    XML_EVOQUE = "xml+evoque"
    EXECLINE = "execline"
    EZHIL = "ezhil"
    FSHARP = "fsharp"
    FSTAR = "fstar"
    FACTOR = "factor"
    FANCY = "fancy"
    FAN = "fan"
    FELIX = "felix"
    FENNEL = "fennel"
    FIFT = "fift"
    FISH = "fish"
    FLATLINE = "flatline"
    FLOSCRIPT = "floscript"
    FORTH = "forth"
    FORTRANFIXED = "fortranfixed"
    FORTRAN = "fortran"
    FOXPRO = "foxpro"
    FREEFEM = "freefem"
    FUNC = "func"
    FUTHARK = "futhark"
    GAP_CONSOLE = "gap-console"
    GAP = "gap"
    GDSCRIPT = "gdscript"
    GLSL = "glsl"
    GSQL = "gsql"
    GAS = "gas"
    GCODE = "gcode"
    GENSHI = "genshi"
    GENSHITEXT = "genshitext"
    POT = "pot"
    GHERKIN = "gherkin"
    GNUPLOT = "gnuplot"
    GO = "go"
    GOLO = "golo"
    GOODDATA_CL = "gooddata-cl"
    GOSU = "gosu"
    GST = "gst"
    GRAPHVIZ = "graphviz"
    GROFF = "groff"
    GROOVY = "groovy"
    HLSL = "hlsl"
    HTML_UL4 = "html+ul4"
    HAML = "haml"
    HTML_HANDLEBARS = "html+handlebars"
    HANDLEBARS = "handlebars"
    HASKELL = "haskell"
    HAXE = "haxe"
    HEXDUMP = "hexdump"
    HSAIL = "hsail"
    HSPEC = "hspec"
    HTML_DJANGO = "html+django"
    HTML_GENSHI = "html+genshi"
    HTML = "html"
    HTML_PHP = "html+php"
    HTML_SMARTY = "html+smarty"
    HTTP = "http"
    HAXEML = "haxeml"
    HYLANG = "hylang"
    HYBRIS = "hybris"
    IDL = "idl"
    ICON = "icon"
    IDRIS = "idris"
    IGOR = "igor"
    INFORM6 = "inform6"
    I6T = "i6t"
    INFORM7 = "inform7"
    INI = "ini"
    IO = "io"
    IOKE = "ioke"
    IRC = "irc"
    ISABELLE = "isabelle"
    J = "j"
    JMESPATH = "jmespath"
    JSLT = "jslt"
    JAGS = "jags"
    JASMIN = "jasmin"
    JAVA = "java"
    JAVASCRIPT_DJANGO = "javascript+django"
    JAVASCRIPT_RUBY = "javascript+ruby"
    JS_GENSHITEXT = "js+genshitext"
    JAVASCRIPT = "javascript"
    JAVASCRIPT_PHP = "javascript+php"
    JAVASCRIPT_SMARTY = "javascript+smarty"
    JS_UL4 = "js+ul4"
    JCL = "jcl"
    JSGF = "jsgf"
    JSONLD = "jsonld"
    JSON = "json"
    JSONNET = "jsonnet"
    JSP = "jsp"
    JLCON = "jlcon"
    JULIA = "julia"
    JUTTLE = "juttle"
    K = "k"
    KAL = "kal"
    KCONFIG = "kconfig"
    KMSG = "kmsg"
    KOKA = "koka"
    KOTLIN = "kotlin"
    KUIN = "kuin"
    LSL = "lsl"
    CSS_LASSO = "css+lasso"
    HTML_LASSO = "html+lasso"
    JAVASCRIPT_LASSO = "javascript+lasso"
    LASSO = "lasso"
    XML_LASSO = "xml+lasso"
    LEAN = "lean"
    LESS = "less"
    LIGHTTPD = "lighttpd"
    LILYPOND = "lilypond"
    LIMBO = "limbo"
    LIQUID = "liquid"
    LITERATE_AGDA = "literate-agda"
    LITERATE_CRYPTOL = "literate-cryptol"
    LITERATE_HASKELL = "literate-haskell"
    LITERATE_IDRIS = "literate-idris"
    LIVESCRIPT = "livescript"
    LLVM = "llvm"
    LLVM_MIR_BODY = "llvm-mir-body"
    LLVM_MIR = "llvm-mir"
    LOGOS = "logos"
    LOGTALK = "logtalk"
    LUA = "lua"
    MCFUNCTION = "mcfunction"
    MCSCHEMA = "mcschema"
    MIME = "mime"
    MIPS = "mips"
    MOOCODE = "moocode"
    DOSCON = "doscon"
    MACAULAY2 = "macaulay2"
    MAKE = "make"
    CSS_MAKO = "css+mako"
    HTML_MAKO = "html+mako"
    JAVASCRIPT_MAKO = "javascript+mako"
    MAKO = "mako"
    XML_MAKO = "xml+mako"
    MAQL = "maql"
    MARKDOWN = "markdown"
    MASK = "mask"
    MASON = "mason"
    MATHEMATICA = "mathematica"
    MATLAB = "matlab"
    MATLABSESSION = "matlabsession"
    MAXIMA = "maxima"
    MESON = "meson"
    MINID = "minid"
    MINISCRIPT = "miniscript"
    MODELICA = "modelica"
    MODULA2 = "modula2"
    TRAC_WIKI = "trac-wiki"
    MONKEY = "monkey"
    MONTE = "monte"
    MOONSCRIPT = "moonscript"
    MOSEL = "mosel"
    CSS_MOZPREPROC = "css+mozpreproc"
    MOZHASHPREPROC = "mozhashpreproc"
    JAVASCRIPT_MOZPREPROC = "javascript+mozpreproc"
    MOZPERCENTPREPROC = "mozpercentpreproc"
    XUL_MOZPREPROC = "xul+mozpreproc"
    MQL = "mql"
    MSCGEN = "mscgen"
    MUPAD = "mupad"
    MXML = "mxml"
    MYSQL = "mysql"
    CSS_MYGHTY = "css+myghty"
    HTML_MYGHTY = "html+myghty"
    JAVASCRIPT_MYGHTY = "javascript+myghty"
    MYGHTY = "myghty"
    XML_MYGHTY = "xml+myghty"
    NCL = "ncl"
    NSIS = "nsis"
    NASM = "nasm"
    OBJDUMP_NASM = "objdump-nasm"
    NEMERLE = "nemerle"
    NESC = "nesc"
    NESTEDTEXT = "nestedtext"
    NEWLISP = "newlisp"
    NEWSPEAK = "newspeak"
    NGINX = "nginx"
    NIMROD = "nimrod"
    NIT = "nit"
    NIXOS = "nixos"
    NODEJSREPL = "nodejsrepl"
    NOTMUCH = "notmuch"
    NUSMV = "nusmv"
    NUMPY = "numpy"
    OBJDUMP = "objdump"
    OBJECTIVE_C = "objective-c"
    OBJECTIVE_C__ = "objective-c++"
    OBJECTIVE_J = "objective-j"
    OCAML = "ocaml"
    OCTAVE = "octave"
    ODIN = "odin"
    OMG_IDL = "omg-idl"
    OOC = "ooc"
    OPA = "opa"
    OPENEDGE = "openedge"
    OUTPUT = "output"
    PACMANCONF = "pacmanconf"
    PAN = "pan"
    PARASAIL = "parasail"
    PAWN = "pawn"
    PEG = "peg"
    PERL6 = "perl6"
    PERL = "perl"
    PHIX = "phix"
    PHP = "php"
    PIG = "pig"
    PIKE = "pike"
    PKGCONFIG = "pkgconfig"
    PLPGSQL = "plpgsql"
    POINTLESS = "pointless"
    PONY = "pony"
    PORTUGOL = "portugol"
    POSTSCRIPT = "postscript"
    PSQL = "psql"
    POSTGRESQL = "postgresql"
    POV = "pov"
    POWERSHELL = "powershell"
    PWSH_SESSION = "pwsh-session"
    PRAAT = "praat"
    PROCFILE = "procfile"
    PROLOG = "prolog"
    PROMQL = "promql"
    PROPERTIES = "properties"
    PROTOBUF = "protobuf"
    PSYSH = "psysh"
    PUG = "pug"
    PUPPET = "puppet"
    PYPYLOG = "pypylog"
    PYTHON2 = "python2"
    PY2TB = "py2tb"
    PYCON = "pycon"
    PYTHON = "python"
    PYTB = "pytb"
    PY_UL4 = "py+ul4"
    QBASIC = "qbasic"
    Q = "q"
    QVTO = "qvto"
    QLIK = "qlik"
    QML = "qml"
    RCONSOLE = "rconsole"
    RNG_COMPACT = "rng-compact"
    SPEC = "spec"
    RACKET = "racket"
    RAGEL_C = "ragel-c"
    RAGEL_CPP = "ragel-cpp"
    RAGEL_D = "ragel-d"
    RAGEL_EM = "ragel-em"
    RAGEL_JAVA = "ragel-java"
    RAGEL = "ragel"
    RAGEL_OBJC = "ragel-objc"
    RAGEL_RUBY = "ragel-ruby"
    RD = "rd"
    REASONML = "reasonml"
    REBOL = "rebol"
    RED = "red"
    REDCODE = "redcode"
    REGISTRY = "registry"
    RESOURCEBUNDLE = "resourcebundle"
    REXX = "rexx"
    RHTML = "rhtml"
    RIDE = "ride"
    RITA = "rita"
    ROBOCONF_GRAPH = "roboconf-graph"
    ROBOCONF_INSTANCES = "roboconf-instances"
    ROBOTFRAMEWORK = "robotframework"
    RQL = "rql"
    RSL = "rsl"
    RESTRUCTUREDTEXT = "restructuredtext"
    TRAFFICSCRIPT = "trafficscript"
    RBCON = "rbcon"
    RUBY = "ruby"
    RUST = "rust"
    SAS = "sas"
    SPLUS = "splus"
    SML = "sml"
    SNBT = "snbt"
    SARL = "sarl"
    SASS = "sass"
    SAVI = "savi"
    SCALA = "scala"
    SCAML = "scaml"
    SCDOC = "scdoc"
    SCHEME = "scheme"
    SCILAB = "scilab"
    SCSS = "scss"
    SED = "sed"
    SHEXC = "shexc"
    SHEN = "shen"
    SIEVE = "sieve"
    SILVER = "silver"
    SINGULARITY = "singularity"
    SLASH = "slash"
    SLIM = "slim"
    SLURM = "slurm"
    SMALI = "smali"
    SMALLTALK = "smalltalk"
    SGF = "sgf"
    SMARTY = "smarty"
    SMITHY = "smithy"
    SNOBOL = "snobol"
    SNOWBALL = "snowball"
    SOLIDITY = "solidity"
    SOPHIA = "sophia"
    SP = "sp"
    DEBSOURCES = "debsources"
    SPARQL = "sparql"
    SPICE = "spice"
    SQL_JINJA = "sql+jinja"
    SQL = "sql"
    SQLITE3 = "sqlite3"
    SQUIDCONF = "squidconf"
    SRCINFO = "srcinfo"
    SSP = "ssp"
    STAN = "stan"
    STATA = "stata"
    SUPERCOLLIDER = "supercollider"
    SWIFT = "swift"
    SWIG = "swig"
    SYSTEMVERILOG = "systemverilog"
    TAP = "tap"
    TNT = "tnt"
    TOML = "toml"
    TADS3 = "tads3"
    TAL = "tal"
    TASM = "tasm"
    TCL = "tcl"
    TCSH = "tcsh"
    TCSHCON = "tcshcon"
    TEA = "tea"
    TEAL = "teal"
    TERATERMMACRO = "teratermmacro"
    TERMCAP = "termcap"
    TERMINFO = "terminfo"
    TERRAFORM = "terraform"
    TEX = "tex"
    TEXT = "text"
    TI = "ti"
    THRIFT = "thrift"
    TID = "tid"
    TLB = "tlb"
    TODOTXT = "todotxt"
    TSQL = "tsql"
    TREETOP = "treetop"
    TURTLE = "turtle"
    HTML_TWIG = "html+twig"
    TWIG = "twig"
    TYPESCRIPT = "typescript"
    TYPOSCRIPTCSSDATA = "typoscriptcssdata"
    TYPOSCRIPTHTMLDATA = "typoscripthtmldata"
    TYPOSCRIPT = "typoscript"
    UL4 = "ul4"
    UCODE = "ucode"
    UNICON = "unicon"
    UNIXCONFIG = "unixconfig"
    URBISCRIPT = "urbiscript"
    USD = "usd"
    VBSCRIPT = "vbscript"
    VCL = "vcl"
    VCLSNIPPETS = "vclsnippets"
    VCTREESTATUS = "vctreestatus"
    VGL = "vgl"
    VALA = "vala"
    ASPX_VB = "aspx-vb"
    VB_NET = "vb.net"
    HTML_VELOCITY = "html+velocity"
    VELOCITY = "velocity"
    XML_VELOCITY = "xml+velocity"
    VERILOG = "verilog"
    VHDL = "vhdl"
    VIM = "vim"
    WDIFF = "wdiff"
    WAST = "wast"
    WEBIDL = "webidl"
    WHILEY = "whiley"
    WOWTOC = "wowtoc"
    WREN = "wren"
    X10 = "x10"
    XML_UL4 = "xml+ul4"
    XQUERY = "xquery"
    XML_DJANGO = "xml+django"
    XML_RUBY = "xml+ruby"
    XML = "xml"
    XML_PHP = "xml+php"
    XML_SMARTY = "xml+smarty"
    XORG_CONF = "xorg.conf"
    XSLT = "xslt"
    XTEND = "xtend"
    EXTEMPORE = "extempore"
    YAML_JINJA = "yaml+jinja"
    YAML = "yaml"
    YANG = "yang"
    ZEEK = "zeek"
    ZEPHIR = "zephir"
    ZIG = "zig"
    ANSYS = "ansys"
    IPYTHON2 = "ipython2"
    IPYTHON3 = "ipython3"
    IPYTHONCONSOLE = "ipythonconsole"


# from pygments.lexers import get_all_lexers, get_lexer_by_name
# for lexer in get_all_lexers():
#     try:
#         name = lexer[1][0]
#         print(f"{name.upper().replace('+', '_').replace('-', '_')} = \"{name}\"", )
#     except:
#         pass



from ..utils import DirectoryTree, Toplevel


class PathView(Toplevel):
    def __init__(self, master, width=150, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.width = round(width)
        
        self.tree = DirectoryTree(self, width=width)
        self.tree.pack()

        self.config(pady=1, padx=1, bg=self.base.theme.border)
        self.overrideredirect(True)
        self.withdraw()

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.configure_bindings()

    def configure_bindings(self):
        self.bind("<FocusOut>", self.hide)
        self.bind("<Escape>", self.hide)
    
    def get_popup_x(self, width):
        return self.winfo_rootx() + int(self.winfo_width() / 2) - int(width / 2)

    def get_popup_y(self):
        return self.winfo_rooty()
    
    def hide(self, *_):
        self.withdraw()
        
    def show(self, e):
        self.update_idletasks()
        w = e.widget
        x = w.winfo_rootx()
        y = w.winfo_rooty() + w.winfo_height()
        
        if not w.path:
            return
        self.tree.change_path(w.path)

        self.geometry(f"+{x}+{y}")
        self.deiconify()
        self.focus_set()



import filetype


class FileType:
    @staticmethod
    def get_file_type(file_path):
        return filetype.guess(file_path)
    
    @staticmethod
    def is_image(file_path):
        return filetype.is_image(file_path)



import tkinter as tk


class Menubutton(tk.Menubutton):
    def __init__(self, master, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.base = master.base



import os
import tkinter as tk

from .frame import Frame
from .tree import Tree


class DirectoryTree(Frame):
    def __init__(self, master, path=None, openfile=None, preview_file=None, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.openfile = openfile
        self.preview_file = preview_file

        self.nodes = {}
        
        self.ignore_dirs = [".git", "__pycache__"]
        self.ignore_exts = [".pyc"]

        self.tree = Tree(self, path, doubleclick=self.openfile, singleclick=self.preview_file, *args, **kwargs)
        self.tree.grid(row=0, column=0, sticky=tk.NSEW)

        self.path = path
        if path:
            self.change_path(path)
        else:
            self.tree.insert('', 0, text='You have not yet opened a file.')
        self.tree.tree.bind("<<TreeviewOpen>>", self.on_treeview_open)
        
    def change_path(self, path):
        self.nodes.clear()
        self.path = path
        if self.path:
            self.tree.clear_tree()
            self.create_root()
        else:
            self.tree.clear_tree()
            self.tree.insert('', 0, text='Empty')
        
    def create_root(self):
        self.update_treeview([(p, os.path.join(self.path, p)) for p in os.listdir(self.path)])

        for path, item in list(self.nodes.items()):
            if not os.path.exists(path):
                self.tree.delete(item)
                self.nodes.pop(path)

    def get_actionset(self):
        return self.actionset
    
    def scandir(self, path):
        entries = []
        for entry in os.scandir(path):
            entries.append((entry.name, entry.path))
        return entries

    def update_treeview(self, entries, parent=""):
        entries.sort(key=lambda x: (not os.path.isdir(x[1]), x[0]))
        for name, path in entries:
            if os.path.isdir(path):
                if name in self.ignore_dirs:
                    continue

                item = self.tree.tree.insert(parent, "end", text=f"  {name}", values=[path, 'directory'], image='foldericon', open=False)
                self.tree.tree.insert(item, "end", text="loading...")
                self.tree.tree.update_idletasks()
            else:
                if name.split(".")[-1] in self.ignore_exts:
                    continue

                #TODO check filetype and get matching icon, cases
                item = self.tree.tree.insert(parent, "end", text=f"  {name}", values=[path, 'file'], image='document')
                self.tree.tree.update_idletasks()

    def on_treeview_open(self, event):
        item_id = self.tree.tree.focus()
        if not item_id:
            return

        path = self.tree.tree.item(item_id, "values")[0]
        if not path or not os.path.isdir(path):
            return

        parent_id = item_id
        self.tree.tree.delete(*self.tree.tree.get_children(parent_id))
        self.update_treeview(self.scandir(path), parent_id)

    def close_directory(self):
        self.change_path(None)
    
    def openfile(self, _):
        if self.openfile:
            self.openfile()
    
    def preview_file(self, *_):
        if self.preview_file:
            self.preview_file()



import tkinter as tk
from ..utils import Frame


#TODO update minimap when scrollbar is used
class Minimap(Frame):
    def __init__(self, master, textw, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.tw = textw
        self.font = ("Arial", 1, "bold")
        self.config(highlightthickness=0, bg=self.base.theme.border)
        
        self.cw = tk.Canvas(self, width=100, highlightthickness=0, **self.base.theme.minimap)
        self.cw.pack(fill=tk.BOTH, expand=True, side=tk.LEFT, padx=(1, 0))

        self.slider_image = tk.PhotoImage(data="""iVBORw0KGgoAAAANSUhEUgAAAG4AAABFCAYAAACrMNMO
        AAAACXBIWXMAAABfAAAAXwEqnu0dAAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAAMBJRE
        FUeJzt0UENwCAAwMAxLajjhwOkz8M+pMmdgiYda5/5kPPeDuAf46KMizIuyrgo46KMizIuyrgo46KMizIuyrgo
        46KMizIuyrgo46KMizIuyrgo46KMizIuyrgo46KMizIuyrgo46KMizIuyrgo46KMizIuyrgo46KMizIuyrgo46
        KMizIuyrgo46KMizIuyrgo46KMizIuyrgo46KMizIuyrgo46KMizIuyrgo46KMizIuyrgo46KMizIu6gNeAwIJ
        26ERewAAAABJRU5ErkJggg==""")

        self.cw.create_image(0, 0, image=self.slider_image, anchor=tk.NW, tag="slider")

        self.y_top_lim = 0
        self._drag_data = {"y": 0, "item": None}
        self.yvalue = 0

        self.cw.tag_bind("slider", "<ButtonPress-1>", self.drag_start)
        self.cw.tag_bind("slider", "<ButtonRelease-1>", self.drag_stop)
        self.cw.tag_bind("slider", "<B1-Motion>", self.drag)

        if textw:
            self.redraw()

    def attach(self, textw):
        self.tw = textw

    def redraw(self):
        self.cw.delete("redrawn")
        self.text = self.tw.get('1.0', tk.END)
        self.cw.create_text(5, 0, text=self.text, anchor=tk.NW, font=self.font, fill="grey", tag="redrawn")

        self.y_bottom_lim = int(self.tw.index(tk.END).split(".")[0]) * 2 + 10
        # self.y_bottom_lim = self.tw.yview()[1] * self.cw.winfo_height()
    
    def redraw_cursor(self):
        self.cw.delete("cursor")
        y = int(self.tw.index(tk.INSERT).split(".")[0]) * 2
        self.cw.create_line(0, y, 100, y, fill="#dc8c34", width=2, tag="cursor")

    def drag_start(self, event):
        self._drag_data["item"] = self.cw.find_closest(event.x, event.y)[0]
        self._drag_data["y"] = event.y

    def drag_stop(self, event):
        self._drag_data["item"] = None
        self._drag_data["y"] = 0

    def drag(self, event):
        item = self._drag_data["item"]
        if item != 1:
            return

        delta_y = event.y - self._drag_data["y"]
        self.cw.move(item, 0, delta_y)
        self._drag_data["y"] = event.y

        self.yvalue = y = self.cw.coords(item)[1]
        if y <= self.y_top_lim:
            self.cw.move("slider", 0, -(y - self.y_top_lim))
        elif y >= self.y_bottom_lim:
            self.cw.move("slider", 0, -(y - self.y_bottom_lim))

        self.tw.yview(int(y / self.cw.winfo_height() * 350))
        self.tw.master.on_scroll()



import tkinter as tk

from .kind import Kind
from ...utils import Frame


class AutoCompleteItem(Frame):
    def __init__(self, master, text, kind=None, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.config(width=400, bg=self.base.theme.autocomplete["background"])
        self.bg, self.fg, self.hbg, self.hfg = self.base.theme.autocomplete.values()

        self.text = text
        self.kind = kind

        self.kindw = Kind(self, self.master.autocomplete_kinds, kind)
        self.textw = tk.Text(self, 
            font=("Consolas", 11), fg=self.fg, bg=self.bg,
            relief=tk.FLAT, highlightthickness=0, width=30, height=1)
        self.textw.insert(tk.END, text)
        self.textw.config(state=tk.DISABLED)

        self.textw.tag_config("term", foreground=self.base.theme.accent)
        
        self.kindw.bind("<Button-1>", self.on_click)
        self.textw.bind("<Button-1>", self.on_click)

        self.bind("<Enter>", self.on_hover)
        self.bind("<Leave>", self.off_hover)

        self.selected = False
        self.hovered = False

        self.grid_columnconfigure(1, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.kindw.grid(row=0, column=0, sticky=tk.NSEW)
        self.textw.grid(row=0, column=1, sticky=tk.NSEW)
    
    def get_text(self):
        return self.text
    
    def get_kind(self):
        return self.kind

    def mark_term(self, term):
        start_pos = self.text.find(term)
        end_pos = start_pos + len(term)
        self.textw.tag_remove("term", 1.0, tk.END)
        self.textw.tag_add("term", f"1.{start_pos}", f"1.{end_pos}")
    
    def on_click(self, *args):
        self.master.choose(self)
    
    def on_hover(self, *args):
        if not self.selected:
            self.kindw.config(bg=self.hbg)
            self.textw.config(bg=self.hbg)
            self.hovered = True

    def off_hover(self, *args):
        if not self.selected:
            self.kindw.config(bg=self.bg)
            self.textw.config(bg=self.bg)
            self.hovered = False
    
    def toggle_selection(self):
        if self.selected:
            self.select()
        else:
            self.deselect()

    def select(self):
        self.kindw.config(bg=self.hbg)
        self.textw.config(bg=self.hbg, fg=self.hfg)
        self.selected = True
    
    def deselect(self):
        self.kindw.config(bg=self.bg)
        self.textw.config(bg=self.bg, fg=self.fg)
        self.selected = False


import tkinter as tk
from tkinter import ttk


class Style(ttk.Style):
    def __init__(self, master, config, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.base = master
        self.config = config
        self.theme = config.theme
        
        self.gen_fileicons()
        self.config_treeview()
        self.config_tree_scrollbar()
    
    def config_tree_scrollbar(self):
        self.element_create("TreeScrollbar.trough", "from", "clam")
        self.element_create("TreeScrollbar.thumb", "from", "clam")

        self.layout("TreeScrollbar", [
            ('TreeScrollbar.trough', {
                'sticky': 'ns',
                'children': [
                    ('TreeScrollbar.thumb', {
                        'unit': '1',
                        'sticky': 'nsew'
                    })
                ]
            })
        ])

        
        bg, highlight = self.theme.scrollbar.values()
        self.configure("TreeScrollbar", gripcount=0, background=bg, troughcolor=bg, bordercolor=bg, lightcolor=bg, darkcolor=bg, arrowsize=14)
        self.map("TreeScrollbar", background=[('pressed', highlight), ('!disabled', self.theme.border)])
        
        self.element_create("EditorScrollbar.trough", "from", "clam")
        self.element_create("EditorScrollbar.thumb", "from", "clam")

        self.layout('EditorScrollbar', [
            ('EditorScrollbar.trough', {
                'sticky': 'nsew',
                'children': [
                    ('EditorScrollbar.thumb', {
                        'sticky': 'nsew'
                    })
                ]
            })
            
        ])
        self.configure("EditorScrollbar", gripcount=0, background=bg, troughcolor=bg, bordercolor=bg, lightcolor=bg, darkcolor=bg)
        self.map("EditorScrollbar", background=[('pressed', highlight), ('!disabled', self.theme.border)])

    def config_treeview(self):
        ## TREENODE CHEVRONS -----
        self.img_tree_close = tk.PhotoImage("img_tree_close", data="""
                iVBORw0KGgoAAAANSUhEUgAAAAsAAAALCAYAAACprHcmAAAACXBIWXMAAA7DAAAOwwHHb6hkAAAAGXRFWHRTb2Z0d
                2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAALBJREFUGJVjYIACExMTfxMTk8WhoaHMDDgAE4zx69ev4////ze4d+
                /ecgcHBxZsihmROXp6emKsrKx7GRgYrvPy8kYdOHDgD07FhDQwoSu+dOnSq9+/fzszMDBofv78eY2npyc7TsUwDez
                s7JEMDAweL1++XAgTx+oRPT09sZ8/fy5nZGTcLiYmFk+Mm6/x8vJGI7sZV2hgKERRTEghigfZ2NgsGRkZLygpKWGE
                LwwAAECxSWJ5KCTqAAAAAElFTkSuQmCC""")
        self.img_tree_open = tk.PhotoImage("img_tree_open", data="""
                iVBORw0KGgoAAAANSUhEUgAAAAsAAAALCAYAAACprHcmAAAACXBIWXMAAA7DAAAOwwHHb6hkAAAAGXRFWHRTb2Z0d
                2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAALxJREFUGJW10D0KwkAQBeA3wd2EiJ23ELbZYJMTKGJhIZ7G81irCB
                4g5TaBNJ7Bzh92Y/GsItsogvjK4Zs3MMC/IkVRjNu2beq6vr1Dxpi+1nqUkFxrrQ/GmP4HeCC5TgAsATyUUseyLAc
                xtNbmSqkdSfHer6QbisiWZJZl2aSqqou1NgewB9Dz3k+bprlK3NItpGm6CCFsYggASYedc3eScxF5hBBOIpLGEABe
                zdGFIcm9iMycc+cvv/pjnkvzViGP6ap9AAAAAElFTkSuQmCC""")
        self.img_tree_empty = tk.PhotoImage("img_tree_empty", data="""
                iVBORw0KGgoAAAANSUhEUgAAAAsAAAALCAYAAACprHcmAAAACXBIWXMAAA7DAAAOwwHHb6hkAAAAGXRFWHRTb2Z0d
                2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAAA5JREFUGJVjYBgFIwgAAAHvAAGLZFZqAAAAAElFTkSuQmCC""")

        self.element_create(
            'Treeitem.nindicator', 'image',  self.img_tree_close,
            ('user1', '!user2', self.img_tree_open), ('user2', self.img_tree_empty), 
            sticky='w', width=20)

        self.configure("Treeview", font=self.config.uifont, rowheight=25, **self.theme.tree)  
        self.map("Treeview", background=[('selected', self.theme.tree["activebackground"])])

        self.layout('Treeview', [('Treeview.treearea', {'sticky': 'nswe'})])
        self.layout('Treeview.Item', [
            ('Treeitem.padding', {
                'sticky': 'nswe',
                'children': [
                    ('Treeitem.nindicator', {
                        'side': 'left', 'sticky': ''
                    }),
                    ('Treeitem.image', {
                        'side': 'left', 'sticky': ''
                    }),
                    ('Treeitem.text', {
                        'side': 'left', 'sticky': ''
                    })
                ]
            })
        ])
    
    def gen_fileicons(self):
        self.document_icn = tk.PhotoImage("document", data="""
        iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAAJ2AAACdgBx6C5rQA
        AABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAADlSURBVDiNpZGxTgJBFEXPW9aGCRTYWht+wyVEEuq1md
        7Exm/A2NjxFUvBD1CQ7JZWlOhXQCNsoYnPajZk3Zks4VaTmfvOO8nAhRF3SJfa2X2XLwi39ZIi74XtzoOA0eIwUZVVQ+cDu
        AHmuTWz+mNUbdGo57HcKiTAc5KVb15AKIU1G4Ux6GMd0grgICIyBX26yw737j5uMZsm2VEBVBUAIeqfbeDLP4PcGmkqujgb
        LyDJjsuLDAJJWwFyax6ainV1L8BX9KX6BZHfr7ZDp93KYBCb9f6nfFUYhoZV+by+MutzLIP5A16TRi/mS3m5AAAAAElFTkS
        uQmCC
        """)

        self.folder_icn = tk.PhotoImage("foldericon", data="""
        iVBORw0KGgoAAAANSUhEUgAAABAAAAAMCAYAAABr5z2BAAAACXBIWXMAAA7DAAAOwwHHb6hkAAAAGXRFWHRTb2Z0d2FyZQB
        3d3cuaW5rc2NhcGUub3Jnm+48GgAAAJBJREFUKJHdzTEKwkAUhOF/loCFRbAVr+IhLAWLCPaW3sFGPIOm1Bt4hxSSEwRs7Z
        UdayErmnROO++bp93htJK0BUa8pxEq1ovZhQ/R/ni+G/LWEjW2y4Stx4NnmUU7l9R6YTxBbFLfb49sGlL4m9ieh84aAA17D
        sCfDLiHdwDqrlpwDTHGAqiA+IONQIW0fAFkySdEGFdeCgAAAABJRU5ErkJggg==""")



import toml, os
from tkinter.font import Font
from types import SimpleNamespace

from .styles import Style

class Config:
    def __init__(self, master, config_file=None, darkmode=True, font=None, uifont=None):
        self.base = master
        self.dir = os.path.dirname(__file__)

        self.stipple = os.path.join(self.dir, 'stipple.xbm')
        if not config_file:
            config_file = os.path.join(self.dir, 'dark.toml') if darkmode else os.path.join(self.dir, 'light.toml')

        self.font = font or Font(family="Consolas", size=11)
        self.uifont = uifont or Font(family="Segoi UI", size=10)
        self.load_from(config_file)
    
    def load_from(self, config_file: str):
        self.theme = SimpleNamespace(**toml.load(config_file))
        self.theme.editor = SimpleNamespace(**self.theme.editor)
        self.theme.diffeditor = SimpleNamespace(**self.theme.diffeditor)
        self.syntax = self.theme.syntax

        self.style = Style(self.base, self)



import tkinter as tk

from .breakpoint import Breakpoint
from ...utils import Canvas, Menubutton


class LineNumbers(Canvas):
    def __init__(self, master, text=None, font=None, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.font = font

        self.fg = self.base.theme.linenumbers["foreground"]
        self.hfg = self.base.theme.linenumbers["activeforeground"]

        self.config(width=65, bd=0, highlightthickness=0, bg=self.base.theme.linenumbers["background"])
        self.text = text

    def attach(self, text):
        self.text = text
    
    def mark_line(self, line):
        dline = self.text.dlineinfo(line)
        
        if not dline:
            return
        
        y = dline[1]
        btn = Menubutton(self, 
            text=">", font=self.font, cursor="hand2", borderwidth=0,
            width=2, height=1, pady=0, padx=0, relief=tk.FLAT, **self.base.theme.linenumbers)
        self.create_window(70, y-2, anchor=tk.NE, window=btn)
    
    def set_bar_width(self, width):
        self.configure(width=width)
        
    def redraw(self, *_):
        self.delete(tk.ALL)

        i = self.text.index("@0,0")
        while True :
            dline = self.text.dlineinfo(i)
            if dline is None: 
                break

            y = dline[1]
            linenum = str(i).split(".")[0]

            curline = self.text.dlineinfo(tk.INSERT)
            cur_y = curline[1] if curline else None

            self.create_text(40, y, anchor=tk.NE, text=linenum, font=self.font, tag=i, fill=self.hfg if y == cur_y else self.fg)
            self.tag_bind(i, "<Button-1>", lambda _, i=i: self.text.select_line(i))

            # TODO drawing breakpoints - need optimisations
            # self.draw_breakpoint(y)
            
            i = self.text.index(f"{i}+1line")
    
    def draw_breakpoint(self, y):
        bp = Breakpoint(self)
        self.create_window(21, y-2, anchor=tk.NE, window=bp)



import tkinter as tk
from tkinter import ttk

from .frame import Frame
from .canvas import Canvas


class Scrollbar(ttk.Scrollbar):
    def set(self, low, high):
        if float(low) <= 0.0 and float(high) >= 1.0:
            self.pack_forget()
        else:
            self.pack(side=tk.RIGHT, fill=tk.Y)
        ttk.Scrollbar.set(self, low, high)


class ScrollableFrame(Frame):
    def __init__(self, master, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        
        self.scrollbar = Scrollbar(self)
        self.scrollbar.pack(side=tk.RIGHT, fill=tk.Y)

        self.canvas = Canvas(self, yscrollcommand=self.scrollbar.set, bg=self.base.theme.background)
        self.canvas.configure(highlightthickness=0)
        self.canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

        self.scrollbar.config(command=self.canvas.yview)

        self.content = Frame(self.canvas, bg=self.base.theme.background)
        self._content = self.canvas.create_window((0, 0), window=self.content, anchor="nw")

        self.content.bind("<Configure>", self._scroll)
        self.canvas.bind("<Configure>", self._configure_canvas)

    def _scroll(self, event):
        self.canvas.configure(scrollregion=self.canvas.bbox("all"))

    def _configure_canvas(self, event):
        canvas_width = event.width
        self.canvas.itemconfig(self._content, width=canvas_width)

    def add(self, content):
        content.pack(in_=self.content)



from ..texteditor import TextEditor

# TODO currently using TextEditor, use Editor instead
class DiffPane(TextEditor):
    def __init__(self, master, *args, **kwargs):
        super().__init__(master, minimalist=True, *args, **kwargs)

    def load_file(self):
        self.text.load_file()
    
    def load_text(self, text):
        self.text.clear_insert(text)



import os
import tkinter as tk
from .pathview import PathView

from ..utils import Frame, Menubutton


class Item(Menubutton):
    def __init__(self, master, path, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.path = path
        self.config(pady=3, padx=3, font=self.base.settings.uifont, **self.base.theme.breadcrumbs)

class BreadCrumbs(Frame):
    def __init__(self, master, path=None, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.config(padx=20, bg=self.base.theme.breadcrumbs["background"])

        self.pathview = PathView(self)
        path = os.path.abspath(path).split('\\')
        for i, item in enumerate(path):
            text = item if item == path[-1] else f"{item} ›"
            self.additem("\\".join(path[:i]), text)

    def additem(self, path, text):
        btn = Item(self, path, text=text)
        btn.bind("<Button-1>", self.pathview.show)
        btn.pack(side=tk.LEFT)



import tkinter as tk
from itertools import chain

from .kinds import Kinds
from .item import AutoCompleteItem

from ...utils import Toplevel


class AutoComplete(Toplevel):
    def __init__(self, master, items=None, active=False, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.autocomplete_kinds = Kinds(self)
        self.config(padx=1, pady=1, bg=self.base.theme.border)
        
        self.active = active
        if not self.active:
            self.withdraw()
        
        self.overrideredirect(True)
        self.wm_attributes("-topmost", True)
        self.grid_columnconfigure(0, weight=1)

        self.menu_items = []
        self.active_items = []
        self.row = 0
        self.selected = 0
        if items:
            # TODO this should be a dict
            self.items = items # [(completion, type), ...]
            self.add_all_items()
            self.refresh_selected()
    
    def update_completions(self):
        self.refresh_geometry()
        self.update_idletasks()
        self.update_all_words()

        term = self.master.get_current_word()

        exact, starts, includes = [], [], []
        for i in self.menu_items:
            if i.get_text() == term:
                exact.append(i)
            elif i.get_text().startswith(term):
                starts.append(i)
            elif term in i.get_text():
                includes.append(i)
        new = list(chain(exact, starts, includes))

        self.hide_all_items()
        if any(new):
            self.show_items(new[:10] if len(new) > 10 else new, term)
        else:
            self.hide()
    
    def move_up(self, *_):
        if self.active:
            self.select(-1)
            return "break"
    
    def move_down(self, *_):
        if self.active:
            self.select(1)
            return "break"

    def add_all_items(self):
        for i in self.items:
            self.add_item(i[0], i[1] if len(i) > 1 else None) 
            
        self.active_items = self.menu_items
        self.refresh_selected()
    
    def update_all_words(self):
        for word in self.master.words:
            if word not in self.get_items_text():
                self.add_item(word, "word")
        
        for word in self.menu_items:
            if word.get_text() not in self.master.words and word.get_kind() == "word":
                self.remove_item(word)

    def add_item(self, text, kind=None):
        new_item = AutoCompleteItem(self, text, kind=kind)
        new_item.grid(row=self.row, sticky=tk.EW)
        
        self.menu_items.append(new_item)

        self.row += 1
    
    def remove_item(self, item):
        a = self.menu_items
        item.grid_forget()
        self.menu_items.remove(item)
        self.row -= 1

    def select(self, delta):
        self.selected += delta
        if self.selected > len(self.active_items) - 1:
            self.selected = 0
        elif self.selected < 0:
            self.selected = len(self.active_items) - 1
        self.refresh_selected()
    
    def reset_selection(self):
        self.selected = 0
        self.refresh_selected()

    def refresh_selected(self):
        for i in self.active_items:
            i.deselect()
        if self.selected < len(self.active_items):
            self.active_items[self.selected].select()
    
    def get_items_text(self):
        return [i.get_text() for i in self.menu_items]
    
    def hide_all_items(self):
        for i in self.menu_items:
            i.grid_forget()
        
        self.active_items = []
        self.row = 1
    
    def show_items(self, items, term):
        self.active_items = items
        for i in items:
            i.grid(row=self.row, sticky=tk.EW)
            self.row += 1

            i.mark_term(term)

        self.reset_selection()
    
    def refresh_geometry(self, *_):
        self.update_idletasks()
        self.geometry("+{}+{}".format(*self.master.cursor_screen_location()))

    def show(self, pos):
        self.active = True
        self.update_idletasks()
        self.geometry("+{}+{}".format(*pos))
        self.deiconify()

    def hide(self, *_):
        self.active = False
        self.withdraw()
        self.reset()
    
    def reset(self):
        self.reset_selection()
    
    def choose(self, this=None, *_):
        if not self.active_items:
            return
        
        if not this:
            this = self.active_items[self.selected]
        
        self.master.confirm_autocomplete(this.get_text())
        self.hide()
        return "break"



from tkinter import ttk


class Scrollbar(ttk.Scrollbar):
    def set(self, low, high):
        if float(low) <= 0.0 and float(high) >= 1.0:
            self.grid_remove()
        else:
            self.grid()
        ttk.Scrollbar.set(self, low, high)



import tkinter as tk


class Kinds:
    def __init__(self, master, *args, **kwargs):
        self.master = master

        self.imethods = tk.PhotoImage(data="""iVBORw0KGgoAAAANSUhEUgAAAA8AAAAPCAYAAAA71pVKAAAACXBIWXMAAA7DAAAOwwHHb6hk
        AAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAAddJREFUKJGNkjFrk3EQxn/3vtEW8QMICoK
        CNZYqDn4BndQ2NIMZRHES0TqILomW9toG7aJiW4cOQkGKoAhNagUHEZwspHQQm9SiDoI4iBqJJNb3/Z9D+0IaE+
        iz3d3z4+7gEZpoLru0OwjkDpAE5p3H5eRAfKHRJ/XF5GRhy46v2y9hNmxITiCNcCqqfeRqj3Z8+w/Oa+kouAlDq
        oL1JfTAm2j2VFd2+YQ3PeykiQwvWse4qjhZA5eyhlxAuL5o+x+oimv2Tn5o+biZGxMoJjSe8AAMuQJUzPg9qFgz
        cM1nNYGyQc8zfb/HW+/7AvcFRvJaKswOvTtWD80MF7vyWprD7IkTmwYqq4QxAchpseog/hf3pU38i5j1AwXDbnv
        IaYOUIRNGdTSph3/mtFgO4UisfkNKO1eBe49HP0y111bTIC8MHlksiPf2d31ufMNrbACk0nvLCY1ngKrzY4PNwJ
        bwZhXBoeC2bQZ4pZ/aAX8rfrAO213Be50bKp5XtZbXzOjSiV/U3gIvu3Xfxw0JM2wcqEUJy2mx4vzYIRfyp2XCI
        j0fW2kLvgfXDDIC0wbnwKZAzoA99NuCG92Zgz8i/wY40mx2eWcY2C3BzgosOM/19Q50zjf6/gEMUNa2RFgfkAAA
        AABJRU5ErkJggg==""")
        self.ivariables = None
        self.ifields = tk.PhotoImage(data="""iVBORw0KGgoAAAANSUhEUgAAAA8AAAANCAYAAAB2HjRBAAAACX
        BIWXMAAA7DAAAOwwHHb6hkAAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAAcBJREFUKJF1k
        jFoUwEQhr97SQzaqODi5KCTRRA3K6iQxA5BnIpZHCwIioJK074XqEZSi2hIK0gpilBBcNStVMX4nkWtFDqVLi5F
        wUlECBrRNu/9Tq0xbW68+7/j/uM3OtSQr8MOlAQ5jPkoYnQ8a89bNbYBqumIGcMYPSYmYwkmm01yghLQMHGrmuE
        pZlqH3ZqO4XADOGgwts24X07bz7X5+QUldv7gLGLY4HsUcdncQIeAu4huQTXV4EH5lP3qaGdWey1kBkjFEc8Mmq
        HoGc/a505Q0deBEFxC+gxmBHkH2CfjoxlLnq+H3qy6W6HB1zru+ZqO4D3wNRT7V1e4CBAHaP6hP5lkRyiuEPLB8
        /UOMS2jH9gjuOeEnKn0Wh1g4IV2rcMAd9L2CSgUX2kkinEVYwK40PWNJ+W8rWxmxWlvVHqt3hRTwGo1Y4/awYE5
        bU1s4Ryw7ABfYklOlwPF2xf99+WX6hryVYj/ZllGDqMvroi84zDRAM8NdLu+ncfU/0HlQKmGuAQMAotE5Ksn7C2
        sJUwy9w0nESVgt2DK4JrBiKAALDhitJK1udZrNsSzGOhoBDcRaaAmuD6WsfnNrPwFMR6ziXNIHQ8AAAAASUVORK
        5CYII=""")
        self.iclasses = tk.PhotoImage(data="""iVBORw0KGgoAAAANSUhEUgAAAA8AAAAPCAYAAAA71pVKAAAAC
        XBIWXMAAA7DAAAOwwHHb6hkAAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAAdFJREFUKJGN
        0k1IVFEYxvH/c+descDgzsaCFhHUeDPBatcmIheCIi2slasZpDZB22hTBrWqhSEhzUdRIAxB1KpshJYRtTAoxgp
        qYYILvVMimDP3vK2KYZSxZ3U4nN/Lw8sRTYlLvf3mkgpmlXBx35iuvW7QJt7fQ60YHTeXVIA7SFG8f7lkZVI74j
        h/5JgzmxVcT+eqN801hpCN1taikbY4LvX2m9ysZBNhrnrXyr0d8vx7oHf1De9VO+ybS+aA22F2YfJb6UBnvJY8B
        XZvBsHg3tyH9VawUsyM4tSX3lOd0Go+eoSsz/luyGt4eWBQsrNhduFZK1wtZsYwTQGLwLysTCr+FT1ANgq8BdLA
        T0lLzdCZ+YIzeN6wkuSUSZcFYGVStbVo5HcQzHbUN+dM/BD25Z80pYFxTMMShwy7KtyAtlQr9LyR7FZzbZs+EcT
        ++hOgD+gSbiDMfZ73WvF20YX39bArdQ5ZwWDEl1uCpk+y44DzHzelXVPAZN3873Ehc9rf7qGh7lrx8MHmu4Sg09
        zGY0EVrGzo+RZs8EnGtGspJRKAGZmumKgAD7csbLss3z/aHXiNr0g3MBsHXobZ6qX/wgArxZ6TMl4AM2G2elHC/
        gCskr1TRm4JBAAAAABJRU5ErkJggg==""")
        self.iinterfaces = tk.PhotoImage(data="""iVBORw0KGgoAAAANSUhEUgAAAA8AAAAICAYAAAAm06XyAA
        AACXBIWXMAAA7DAAAOwwHHb6hkAAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAAQRJREFUG
        JWFkD9LgnEAhJ/7qbODQw211JLRVlMEofYFgqIhglpsqSnfCJeaChRXgz5Aa30A/1ARCbnXJ7Aai5rC95oEkcAb
        D4577sQYFbtOpb+IMLuIKeBJCQ4qq3pJAkRNzwoylTzPSB4Opz+pI5YtIsf0QmDDfTrRnZdUantfpg4EzG21oPV
        BsNT2gsx9KiZ7vqaPgR+1fYHJKmr5DZgcKrsGvgEEc8B7Ja+tYZqjhudDoBPGbf5PyYAAkoJTwyUQEDfVnLZHsB
        /KDU8MY/fNjqApgJOGZ34hUyvQHT3suOUrmxUFysT0HNjEFBNiUeMQz9pO/sSUJPYM08BjLA5rOb3+AQBaXk0WW
        bcyAAAAAElFTkSuQmCC""")
        self.imodules = tk.PhotoImage(data="""iVBORw0KGgoAAAANSUhEUgAAAA8AAAAPCAYAAAA71pVKAAAAC
        XBIWXMAAA7DAAAOwwHHb6hkAAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAAXtJREFUKJGN
        0TuLE2EYxfH/eZLIikkweEFb2xQWostMOhUCUwha+BG0007wEwh22ilabmNpF1CbZCbeECzS2iqsMrmJu2bzvBa
        OQkJMfMr3nB8ceEVxnU7nSLVavQ9cb7VaJ1i6NE13gWfT6fROu93+DmB/wlqt9gC4FEK4uQwBJN0ALha9328Ag8
        Hg0HA4HJnZ5SiK0lUYoN/vt9z9xWg0Opokyb4B5Hl+HNiazWaf/gUBinyrXq8fW5gNYGZah5dzAyiVSqeAAzPLN
        +AcOAghnAZQmqbngYeSduM4vrIOA2RZ9hw4CdwqA2+Bubuf3QQB3P2upI/AG5O0Lem9md37H1z03gEXBNDtds+Z
        2WtJ9TiOf6yZfDiEMDaz7SiKPhiApM9A2d0bGyY3gPJ8Pv8CS1/l7mEDXsgNYDwefwP2KpXKmXW4yPcajcbXvzh
        Jkn1JO+7+JMuyq6tgr9e75u5PJe00m82fC7Mnk8lt4FUI4fEqLOkR8LLoAfALta2R2TiY6xMAAAAASUVORK5CYI
        I=""")
        self.iproperties = tk.PhotoImage(data="""iVBORw0KGgoAAAANSUhEUgAAAA8AAAAPCAYAAAA71pVKAA
        AACXBIWXMAAA7DAAAOwwHHb6hkAAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAAdFJREFUK
        JGN0r9rFEEUB/Dvd+8uxzUhoBCCEkhtEbTbHVdZIiQQfySKWmkjWtkIWtj4D1imsDQIRlCDoBbauOLNzR5aGrBV
        EY0g2gQCtzfztTALinenrxt4n++8Nwwxorrd7ni/379O8gyAKQDvAdwyxqwAQDQCTnrvLcndJBcBnAewC0Be9Qz
        F3vubJJ8aYy6GEGYkrZBcMMZsWGv3WWsvcxDM83xibGzsQ6/X29NoNPaTfExyPkmS1+12ezaKoucAfgy8udVqTZ
        P8lGXZVlmW7yR9BLBYFMUBks9IXgEwM2zsr5Km8jyvZ1n2rSzLuRDCUgjhhaRLkt4C2ByIvfezANBsNicBoAqQt
        Jym6RNJ1wCs/7VzURRHQgh3oyg6Gcdxx1o7XZbl5yzL+t1ud9x7fwPA0RBC/Ad2zs1JWvsNZiQfAQgANgHsBfCQ
        5NUkSb7XK2itzSStSToVx3GnCgJwnOSG936iVqt9SZJkuzLcgQdJrocQzqZp+tI5d1jS/eo85FHBTqdzCMADAKe
        NMa+cc/OS7lSjD4MAUAdwG8C5HbggaRXAUhzHxSgI/PqeLUnbzrllSashhBPGmH9CAKhLukDynqQtAMfSNH3zPx
        AAfgL2/u9cQzl88QAAAABJRU5ErkJggg==""")
        self.ikeywords = tk.PhotoImage(data="""iVBORw0KGgoAAAANSUhEUgAAAA8AAAAPCAYAAAA71pVKAAAA
        CXBIWXMAAA7DAAAOwwHHb6hkAAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAAPRJREFUKJG
        lkrFKxEAURc8bHltbWqytYG0TYiC2VnZ+hK3iByhouX7HfkcmhAgigpXFgt02gpXETObZiEVMWI23PecOl+HBPy
        Le+5cxaGaLLMsWY1yBHRE5NLO3Hrtwzm0BFEVxJCI3Q2XM7BIIPbYLPAOo6n2M8exHWUROxmaFEJ4AkiRZA+sxb
        1KkLMvrTVKM8SHLsmVVVfMY4ynwnqbplZrZbFPZOacATdOIqs7MrP8/E2Z774+HQAjhMc/zFUBVVfOu6/b7jgLn
        Q2VVvQVWXw/ticigNzlS1/X2b+UQwkeapq/f69q2Hb3tgdwBB39ZN5pP4uJac+7GJRAAAAAASUVORK5CYII=""")
        self.iwords = tk.PhotoImage(data="""iVBORw0KGgoAAAANSUhEUgAAAA8AAAAICAYAAAAm06XyAAAACXB
        IWXMAAA7DAAAOwwHHb6hkAAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jnm+48GgAAATBJREFUGJWFjy
        FLQ2EYhc95t31JEQYiGExaLBODYbvJaFBEGBqHGNW4n6BFBUGw2ZUJCoaxMGT3Xm+1LwkLgrILwzLZdwy7wTD0i
        ec9z4EXmEAcx40wDPcn3X5j/xX+gmEYLgA4JLkE4KFSqdzEcdwA0Jc0A0CSzoMgSNrt9pRzrg5gRVLXSNZIfki6
        AnDa6XRKGBsbku5IPpNsJkky55x7AjAv6ZLkV344HJ4VCoUdM9uUNMrlcsuZfBEEwS0ARFG0670/BrDY6/XWq9X
        qCEAr75xrABgAeASwKilHElmGbIgkpwF8ZuL45yiKUgCBpG+SLwCOSG5LmvXebwEomdk9yTVJife+ViwWm/1+f9
        tI1iW1ssIrgIGkd0lmZm9mdk1yr1wudwFUzewkTdOU5MEPhImOHsSTJnYAAAAASUVORK5CYII=""")



import re
import threading
import tkinter as tk

from .pane import DiffPane
from .differ import Differ
from ..editor import BaseEditor


class DiffEditor(BaseEditor):
    def __init__(self, master, path1, path2, language=None, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.border)
        self.grid_columnconfigure(0, weight=1)
        self.grid_columnconfigure(1, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.path1 = path1
        self.path2 = path2
        self.editable = True

        self.lhs_data = []
        self.rhs_data = []

        self.lhs_last_line = 0
        self.rhs_last_line = 0

        self.lhs = DiffPane(self, path1, language=language)
        self.lhs.grid(row=0, column=0, sticky=tk.NSEW, padx=(0, 1))

        self.rhs = DiffPane(self, path2, language=language)
        self.rhs.grid(row=0, column=1, sticky=tk.NSEW)

        self.left = self.lhs.text
        self.right = self.text = self.rhs.text

        self.lhs.scrollbar['command'] = self.on_scrollbar
        self.rhs.scrollbar['command'] = self.on_scrollbar
        self.left['yscrollcommand'] = self.on_textscroll
        self.right['yscrollcommand'] = self.on_textscroll

        self.stipple = self.base.settings.stipple

        self.left.tag_config("addition", background=self.base.theme.diffeditor.notexist, bgstipple=f"@{self.stipple}")
        self.left.tag_config("removal", background=self.base.theme.diffeditor.removed)
        self.left.tag_config("removedword", background=self.base.theme.diffeditor.removedword)
        
        self.right.tag_config("addition", background=self.base.theme.diffeditor.added)
        self.right.tag_config("removal", background=self.base.theme.diffeditor.notexist, bgstipple=f"@{self.stipple}")
        self.right.tag_config("addedword", background=self.base.theme.diffeditor.addedword)

        self.differ = Differ(self)

        if path1 and path2:
            with open(self.path1, 'r') as f:
                self.lhs_data = f.read()
            with open(self.path2, 'r') as f:
                self.rhs_data = f.read()
    
    def on_scrollbar(self, *args):
        self.left.yview(*args)
        self.lhs.on_scroll()

        self.right.yview(*args)
        self.rhs.on_scroll()

    def on_textscroll(self, *args):
        self.lhs.scrollbar.set(*args)
        self.rhs.scrollbar.set(*args)
        self.on_scrollbar('moveto', args[0])
    
    def run_show_diff(self):
        threading.Thread(target=self.show_diff).start()
    
    def show_diff_text(self, lhs, rhs):
        self.lhs_data = lhs
        self.rhs_data = rhs

        self.show_diff()

    def show_diff(self):
        self.left.set_active(True)
        self.lhs.clear()
        self.rhs.clear()

        lhs_lines = [line+"\n" for line in self.lhs_data.split('\n')]
        rhs_lines = [line+"\n" for line in self.rhs_data.split('\n')]
        
        self.diff = list(self.differ.get_diff(lhs_lines, rhs_lines))
        for i, line in enumerate(self.diff):
            marker = line[0]
            content = line[2:]

            match marker:
                case " ":
                    # line is same in both
                    self.left.write(content)
                    self.right.write(content)

                case "-":
                    # line is only on the left
                    self.left.write(content, "removal")
                    self.left.newline("addition")

                case "+":
                    # line is only on the right
                    self.right.write(content, "addition")
                    self.left.newline("addition")

                # case "?":
                #     # the above line has changes
                #     if matches := re.finditer(r'\++', content):
                #         self.left.delete(str(float(self.rhs_last_line+1)), str(float(int(float(self.left.index(tk.INSERT))))))
                #         for match in matches:
                #             start = f"{self.rhs_last_line}.{match.start()}"
                #             end = f"{self.rhs_last_line}.{match.end()}"
                #             self.right.tag_add("addedword", start, end)

                #     if matches := re.finditer(r'-+', content):
                #         self.right.delete(str(float(self.lhs_last_line+1)), str(float(int(float(self.right.index(tk.INSERT))))))
                #         for match in matches:
                #             start = f"{self.lhs_last_line}.{match.start()}"
                #             end = f"{self.lhs_last_line}.{match.end()}"
                #             self.left.tag_add("removedword", start, end)
                    
            self.left.update()
            self.right.update()
        
        self.left.highlighter.highlight()
        self.right.highlighter.highlight()

        # Add extra empty lines at the bottom if one side has fewer lines
        lhs_line_count = int(float(self.left.index(tk.END))) - 1
        rhs_line_count = int(float(self.right.index(tk.END))) - 1
        if lhs_line_count > rhs_line_count:
            extra_newlines = lhs_line_count - rhs_line_count
            for _ in range(extra_newlines):
                self.right.newline()
        elif rhs_line_count > lhs_line_count:
            extra_newlines = rhs_line_count - lhs_line_count
            for _ in range(extra_newlines):
                self.left.newline()
        
        self.left.set_active(False)



__version__ = '0.25.6'
__version_info__ = tuple([ int(num) for num in __version__.split('.')])

__all__ = ["Editor", "get_editor", "DiffEditor", "ImageViewer", "TextEditor", "Config", "Languages"]


import os
import tkinter as tk
from tkinter.font import Font

from .config import Config
from .languages import Languages
from .utils import FileType, Frame

from .breadcrumbs import BreadCrumbs
from .diffeditor import DiffEditor
from .imageviewer import ImageViewer
from .texteditor import TextEditor

def get_editor(base, path=None, path2=None, diff=False, language=None):
    "picks the right editor for the given values"
    if diff:
        return DiffEditor(base, path, path2, language=language)
    
    if path and os.path.isfile(path):
        if FileType.is_image(path):
            return ImageViewer(base, path)
        
        return TextEditor(base, path, language=language)
    
    return TextEditor(base, language=language)


class Editor(Frame):
    """
    Editor class
    Picks the right editor based on the path, path2, diff values passed. Supports showing diff, images, text files. 
    If nothing is passed, empty text editor is opened. 
    
    Attributes
    ----------
    path : str
        path of the file to be opened
    path2 : str
        path of file to be opened in diff, required if diff=True is passed
    diff : bool
        whether this is to be opened in diff editor
    language : str
        Use the `Languages` enum provided (eg. Languages.PYTHON, Languages.TYPESCRIPT)
        This is given priority while picking suitable highlighter. If not passed, guesses from file extension.
    dark_mode : str
        Sets the editor theme to cupcake dark if True, or cupcake light by default
        This is ignored if custom config_file path is passed
    config_file : str
        path to the custom config (TOML) file, uses theme defaults if not passed
    showpath : bool
        whether to show the breadcrumbs for editor or not
    font : str | Font
        Font used in line numbers, text editor, autocomplete. defaults to Consolas(11)
    uifont : str | Font
        Font used for other UI components (breadcrumbs, trees)
    preview_file_callback : function(path)
        called when files in breadcrumbs-pathview are single clicked. MUST take an argument (path)
    open_file_callback : function(path)
        called when files in breadcrumbs-pathview are double clicked. MUST take an argument (path)

    NOTE: All the *tk.Text* methods are available under *Editor.content* (eg. Editor.content.insert, Editor.content.get)

    Methods
    -------
    save(path: str=None)
        If the content is editable writes to the specified path.
    focus()
        Gives focus to the content.
    """
    def __init__(self, master, 
                 path: str=None, path2: str=None, diff: bool=False, language: str=None,
                 darkmode=True, config_file: str=None, showpath: bool=True, 
                 font: str|Font=None, uifont: str|Font=None,
                 preview_file_callback=None, open_file_callback=None,
                 *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)

        self.path = path
        self.path2 = path2
        self.diff = diff
        self.showpath = showpath
        self.darkmode = darkmode
        self.config_file = config_file
        self.preview_file_callback = preview_file_callback
        self.open_file_callback = open_file_callback

        self.settings = Config(self, config_file, darkmode, font, uifont)
        self.theme = self.settings.theme

        self.config(bg=self.theme.border)
        self.grid_columnconfigure(0, weight=1)

        self.content = get_editor(self, path, path2, diff, language)
        self.filename = os.path.basename(self.path) if path else None
        if path and self.showpath and not diff:
            self.breadcrumbs = BreadCrumbs(self, path)
            self.grid_rowconfigure(1, weight=1)  
            self.breadcrumbs.grid(row=0, column=0, sticky=tk.EW, pady=(0, 1))
            self.content.grid(row=1, column=0, sticky=tk.NSEW)
        else:
            self.grid_rowconfigure(0, weight=1)
            self.content.grid(row=0, column=0, sticky=tk.NSEW)
    
    def save(self, path: str=None) -> None:
        self.content.save(path)
    
    def focus(self) -> None:
        self.content.focus()



import difflib


class Differ(difflib.Differ):
    def __init__(self, master, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.master = master
        self.base = master.base

    def get_diff(self, lhs, rhs):
        return self.compare(lhs, rhs)



"""
Various types of widgets/functions used across the editor
"""

from .directorytree import DirectoryTree
from .canvas import Canvas
from .frame import Frame
from .filetype import FileType
from .label import Label
from .scrollbar import Scrollbar
from .scrollableframe import ScrollableFrame
from .tree import Tree
from .menubutton import Menubutton
from .toplevel import Toplevel
from .text import Text

@staticmethod
def clamp(value, min_val, max_val):
    return min(max(min_val, value), max_val)



import tkinter as tk
from ...utils import Label


class Kind(Label):
    def __init__(self, master, kinds, kind="text", *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.base = master.base
        
        self.kinds = kinds
        self.kind = kind

        self.image = None

        self.config(bg=self.base.theme.autocomplete["background"])
        self.config_image()

    def config_image(self):
        match self.kind:
            case "method":
                self.image = self.kinds.imethods
            case "variable":
                self.image = self.kinds.ivariables
            case "field":
                self.image = self.kinds.ifields
            case "class":
                self.image = self.kinds.iclasses
            case "interface":
                self.image = self.kinds.iinterfaces
            case "module":
                self.image = self.kinds.imodules
            case "property":
                self.image = self.kinds.iproperties
            case "keyword":
                self.image = self.kinds.ikeywords
            case _:
                self.image = self.kinds.iwords
        self.config(image=self.image)


import tkinter as tk

from ..utils import Scrollbar
from ..editor import BaseEditor

from .minimap import Minimap
from .linenumbers import LineNumbers
from .text import Text


class TextEditor(BaseEditor):
    def __init__(self, master, path=None, language=None, font=None, minimalist=False, *args, **kwargs):
        super().__init__(master, path, *args, **kwargs)
        self.font = font or self.base.settings.font
        self.minimalist = minimalist
        self.language = language
        
        self.rowconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)

        self.text = Text(self, path=self.path, minimalist=minimalist, language=language)
        self.linenumbers = LineNumbers(self, self.text, self.font)
        self.scrollbar = Scrollbar(self, orient=tk.VERTICAL, command=self.text.yview, style="EditorScrollbar")
        
        self.text.config(font=self.font)
        self.text.configure(yscrollcommand=self.scrollbar.set)
        
        if not self.minimalist:
            self.minimap = Minimap(self, self.text)
            self.minimap.grid(row=0, column=2, sticky=tk.NS)
        
        self.linenumbers.grid(row=0, column=0, sticky=tk.NS)
        self.text.grid(row=0, column=1, sticky=tk.NSEW)
        self.scrollbar.grid(row=0, column=3, sticky=tk.NS)

        self.text.bind("<<Change>>", self.on_change)
        self.text.bind("<<Scroll>>", self.on_scroll)

    def on_change(self, *_):
        self.text.refresh()
        self.linenumbers.redraw()
        if not self.minimalist:
            self.minimap.redraw_cursor()

    def on_scroll(self, *_):
        self.linenumbers.redraw()
        if not self.minimalist:
            self.minimap.redraw()

    def unsupported_file(self):
        self.text.highlighter.lexer = None
        self.text.show_unsupported_dialog()
        self.linenumbers.grid_remove()
        self.scrollbar.grid_remove()
        self.editable = False

    def focus(self):
        self.text.focus()
        self.on_change()

    def set_fontsize(self, size):
        self.font.configure(size=size)
        self.linenumbers.set_bar_width(size * 3)
        self.on_change()
    
    def save(self, path=None):
        if self.editable:
            self.text.save_file(path)
    
    def cut(self, *_):
        if self.editable:
            self.text.cut()
    
    def copy(self, *_):
        if self.editable:
            self.text.copy()
        
    def paste(self, *_):
        if self.editable:
            self.text.paste()
    
    def write(self, *args, **kwargs):
        if self.editable:
            self.text.write(*args, **kwargs)

    def insert(self, *args, **kwargs):
        if self.editable:
            self.text.insert(*args, **kwargs)

    def get(self, *args, **kwargs):
        if self.editable:
            self.text.get(*args, **kwargs)
    
    def clear(self):
        self.delete("1.0", tk.END)

    def delete(self, *args, **kwargs):
        if self.editable:
            self.text.delete(*args, **kwargs)
    
    def mark_set(self, *args, **kwargs):
        if self.editable:
            self.text.mark_set(*args, **kwargs)

    def compare(self, *args, **kwargs):
        return self.text.compare(*args, **kwargs)

    def dlineinfo(self, index):
        return self.text.dlineinfo(index)

    def edit_modified(self, arg=None):
        return self.text.edit_modified(arg)

    def edit_redo(self):
        if self.editable:
            self.text.edit_redo()
    
    def edit_reset(self):
        if self.editable:
         self.text.edit_reset()

    def edit_separator(self):
        if self.editable:
            self.text.edit_separator()

    def edit_undo(self):
        if self.editable:
            self.text.edit_undo()

    def image_create(self, index, **kwargs):
        if self.editable:
            return self.text.image_create(index, **kwargs)

    def image_cget(self, index, option):
        return self.text.image_cget(index, option)

    def image_configure(self, index, **kwargs):
        if self.editable:
            return self.text.image_configure(index, **kwargs)
    
    def image_names(self):
        return self.text.image_names()

    def index(self, i):
        return self.text.index(i)

    def mark_gravity(self, mark, gravity=None):
        return self.text.mark_gravity(mark, gravity)

    def mark_names(self):
        return self.text.mark_names()
        
    def mark_next(self, index):
        return self.text.mark_next(index)

    def mark_previous(self, index):
        return self.text.mark_previous(index)

    def mark_set(self, mark, index):
        if self.editable:
            self.text.mark_set(mark, index)

    def mark_unset(self, mark):
        if self.editable:
            self.text.mark_unset(mark)

    def scan_dragto(self, x, y):
        self.text.scan_dragto(x, y)

    def scan_mark(self, x, y):
        self.text.scan_mark(x, y)

    def search(self, pattern, index, **kwargs):
        return self.text.search(pattern, index, **kwargs)

    def see(self, index):
        self.text.see(index)

    def tag_add(self, tagName, index1, index2=None):
        if self.editable:
            self.text.tag_add(tagName, index1, index2)

    def tag_bind(self, tagName, sequence, func, add=None):
        self.text.tag_bind(tagName, sequence, func, add)

    def tag_cget(self, tagName, option):
        return self.text.tag_cget(tagName, option)

    def tag_config(self, tagName, **kwargs):
        if self.editable:
            self.text.tag_config(tagName, **kwargs)

    def tag_names(self, index=None):
        return self.text.tag_names(index)

    def tag_nextrange(self, tagName, index1, index2=None):
        return self.text.tag_nextrange(tagName, index1, index2)

    def tag_prevrange(self, tagName, index1, index2=None):
        return self.text.tag_prevrange(tagName, index1, index2)

    def tag_raise(self, tagName, aboveThis=None):
        if self.editable:
            self.text.tag_raise(tagName, aboveThis)

    def tag_ranges(self, tagName):
        return self.text.tag_ranges(tagName)

    def tag_remove(self, tagName, index1, index2=None):
        if self.editable:
            self.text.tag_remove(tagName, index1, index2)

    def tag_unbind(self, tagName, sequence, funcid=None):
        self.text.tag_unbind(tagName, sequence, funcid)

    def window_cget(self, index, option):
        return self.text.window_cget(index, option)

    def window_configure(self, index, **kwargs):
        if self.editable:
            self.text.window_configure(index, **kwargs)

    def window_create(self, index, **kwargs):
        if self.editable:
            self.text.window_create(index, **kwargs)

    def window_names(self):
        return self.text.window_names()

    def xview_moveto(self, fraction):
        self.text.xview_moveto(fraction)

    def xview_scroll(self, n, what):
        self.text.xview_scroll(n, what)

    def yview_moveto(self, fraction):
        self.text.yview_moveto(fraction)

    def yview_scroll(self, n, what):
        self.text.yview_scroll(n, what)



import tkinter as tk


class Text(tk.Text):
    """
    normal text with reference to base
    """
    def __init__(self, master, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.base = master.base
